#include <stdio.h>
#include <stdlib.h>
#include "matrix.h"
#include "fill.h"


#define BACKTONE	255
#define OBJTONE		0
#define FILLTONE	10

#define UP   	0
#define DOWN	1
#define E1  	2
#define E2  	3
#define NPARAM 	4

/** Variables for STACK **/

typedef int typexy;
typedef unsigned char uchar;

#define STACK_SIZE 3000

POS stack[ STACK_SIZE ];
int pos_stack = -1; 

/********** Funtiones for the stack ******************/
int push( int a, int b )
{
	int pone;
 
	pone = pos_stack;
	pone++;
	if ( pone >= STACK_SIZE ){
		fprintf( stderr, "Stack is Full %d!\n", pone );
		exit(1);
	}
	stack[pone].a = a;
	stack[pone].b = b;
	pos_stack = pone;
	return( 1 );      /** ok! **/
}

int pop( int *a, int *b )
{
	if ( pos_stack<0 ) return(0);   /** Error **/
	*a = stack[pos_stack].a;
	*b = stack[pos_stack].b;
	pos_stack--;
	return( 1 );    /** Ok! **/
}

int empty_stack( void )
{
	if( pos_stack<0 ) return( 0 ); /** It's empty **/
	return( 1 );                    /** There is something **/
}

void clean_stack( void )
{
	pos_stack = -1;
}

int fill ( MATRIX *IMG, int x, int y, byte tono )
{
	byte *pimg;

	pimg = (byte *)IMG->ptr[y] + x;
	while ( *pimg == OBJTONE ) {
		*pimg++ = tono;
		x++;
	}
	return( --x );
}

int left ( MATRIX *IMG, int x, int y ) 
{
	byte *pimg;

	pimg = (byte *)IMG->ptr[y] + x;
	while ( *pimg == OBJTONE ) {
		--x; 
		pimg--;
	}
	return ( ++x );
}

int link( MATRIX *IMG, int x, int y, int *par  )
{
	int above=0, below=0;
	int e1, e2 = -1;
	byte **img;

	img = (byte **)IMG->ptr;

	if ( img[y-1][x-1]==OBJTONE ) {
		above++;
		e1 = left( IMG, x-1, y-1 );
		// if ( e1<0 ) return( 0 );
	}
	if ( img[y+1][x-1]==OBJTONE ) {
		below++;
		e2 = left( IMG, x-1, y+1 );
		// if ( e2<0 ) return( 0 );
	}
	while ( img[y][x] == OBJTONE ) {
		if ( img[y-1][x-1]==BACKTONE && img[y-1][x]==OBJTONE ) {
			if ( above ) push ( e1, y-1 );
			above++;
			e1 = left( IMG, x, y-1 );
			// if ( e1<0 ) return( 0 );
		}
		if ( img[y+1][x-1]==BACKTONE && img[y+1][x]==OBJTONE ) {
			if ( below ) push ( e2, y+1 );
			below++;
			e2 = left( IMG, x, y+1 );
			// if ( e2<0 ) return( 0 );
		}
		x++;
		// if ( x==IMG->cols ) return( 0 );
	}
	if ( img[y-1][x-1]==BACKTONE && img[y-1][x]==OBJTONE ){
		if ( above ) push ( e1, y-1 );
		above++;
		e1 = left( IMG, x, y-1 );
		// if ( e1<0 ) return( 0 );
	}
	if ( img[y+1][x-1]==BACKTONE && img[y+1][x]==OBJTONE ){
		if ( below ) push ( e2, y+1 );
		below++;
		e2 = left( IMG, x, y+1 );
		// if ( e2<0 ) return( 0 );
	}
	par[UP] = above;
	par[DOWN] = below;
	par[E1] = e1;
	par[E2] = e2;
	return( 1 );
}


int do_fill( MATRIX *IMG, int *val, byte tono )
{
	int x, y;
	int param[ NPARAM ];
	int pnext, dir;
	int p1x, p1y, p2x, p2y, v;
	int area = 0;
	byte **img;

	p1x = IMG->cols;
	p2x = 0;

	p1y =IMG->rows;
	p2y = 0;

	img = (byte **)IMG->ptr;
	dir = UP;
	while( pop( &x, &y ) ){
		
		if ( x < p1x ) p1x = x;
		else if ( x > p2x ) p2x = x;
		if ( y < p1y ) p1y = y;
		else if ( y > p2y ) p2y = y;

		for(;;){
			if ( !link( IMG, x, y, param ) )
				return(0);
			if ( !param[UP] && !param[DOWN] ){
				if ( img[y][x]==OBJTONE ) {
					v = fill( IMG, x, y, tono );
					area += v - x + 1;
					if ( v > p2x ) p2x = v;
				}
				break;
			}

			if ( !param[UP] ) dir=DOWN;
			else if ( !param[DOWN] ) dir=UP;

			if ( param[UP] && param[DOWN] ) {
				if( dir==UP ) push ( param[ E1+DOWN ], y+1 );
				else  push ( param[ E1+UP ], y-1 );
			}
			pnext = param[ E1+dir ];

			v = fill( IMG, x, y, tono );
			area += v - x + 1;
			if ( v > p2x ) p2x = v;

			x = pnext;
			y = (dir==UP) ? y-1 : y+1;

			if ( x < p1x ) p1x = x;
			else if ( x > p2x ) p2x = x;
			if ( y < p1y ) p1y = y;
			else if ( y > p2y ) p2y = y;
		}
	}
	val[INICIOX] = p1x;
	val[INICIOY] = p1y;
	val[FINX] = p2x;
	val[FINY] = p2y;
	val[AREA] = area;
	return( 0 );
}
